/************************************************************************************
 * CToast.h
 *
 * CheeseToast by Jim Bumgardner
 *
 * 2/28/95
 *		Updated to compile using newer Think C / Metrowerks compiler
 *
 ************************************************************************************/

#include <stdarg.h>
#include "QDOffscreen.h"
#if __option(profile)			// 6/15 Optional profiling support
#include <profile.h>
#endif

#include "M-Assets.h"

#ifndef THINK_C
#define OK		1
#define Cancel	2
#endif

/*
 * Constants
 */

#define DEBUGGING		1		// Enable debugging key and debugging sprites
#define DEBSPRITES		0
#define MaxSprites		3000	// Maximum simultaneous sprites (incl. bullets/shards)
// #define MaxSaveMaps		100		// Maximum Large Sprites
#define MaxScoreRecords	10		// Number of top scores shown
#define ColorMapSize	1024L	// Size of 32x32 8-bit sprite map
#define MaskMapSize		128L	// Size of 32x32 1-bit MaskMap
#define BulletMapSize	4L		// Size of 2x2 8-bit bullet map
#define IconWidth		32		// Width of Icon
#define IconHeight		32		// Height of Icon
#define LargeIconWidth	48		// Width of Icon
#define LargeIconHeight	48		// Height of Icon
#define SafeWidth		48
#define SafeHeight		48
#define BulletWidth		2		// Width/Height of Bullet
#define StartupPICT		128		// Starup Picture Resource ID
// #define RadarPICT		129
#define ScoreTitlePICT	130		// "CheezeWhizzes" title
#define SmallLogoPICT	131		// Small Logo Picture ID (Attract Mode)
#define RadarMaskPICT	132
#define NbrAsteroids	3		// Number basic asteroids in table
#define MaxAngles		32		// Number of vectors in vector table
#define MaxShieldPower	100		// Maximum Shield Power

#define AsterPoints		100		// Points for zapping an asteroid
#define SaucerPoints	500		// Points for zapping a saucer
#define CroissantPoints	600		// Points for zapping a Croissant
#define CoffeePotPoints	600		// Points for zapping a CoffeePot
#define CubePoints		750		// Points for zapping a cube
#define MuffinPoints	750
#define BeaterPoints	750
#define MixerPoints		750
#define LevelBonusScore	1000	// Points for advancing to next level
#define FreeShipPoints	50000L	// Points required for free teapot

#define StatusHeight	12

#define ShieldIncrease	4

#define NormalYummyChance 		400L	// was 1000
#define CreamerChance			1800L	// once per minute	
#define NormalToasterChance		1040L
#define MaxBadGuyChance			650		// 1 per 25 seconds
#define MinBadGuyChance			133		// 1 every five seconds

#define YummyLifeSpan			200L	// was 100
#define CreamerLifeSpan			150L
#define ToasterLifeSpan			650L
#define CroissantLifeSpan		400
#define CubeLifeSpan			780
#define QuakeLifeSpan			60
#define PhotonLifeSpan			35
#define BulletLifeSpan			35

#define MedSparkLifeSpan		40
#define SmallSparkLifeSpan		25
#define LargeSparkLifeSpan		65

#define BadGuyLevelDecrement	50
#define MinAsteroidsPerLevel	9		// was 3
#define ExtraAsteroidsPerLevel	1		// was .5
#define MaxAsteroids			30		// was 20
#define NewLevelIncrement		10000L

#define TopScoreResType			'tScr'
#define PrefsResType			'prfs'

/*
 * Function Macros
 */
 
#define AniFrameIndex(i)	(i << 10)	// Multiply by 1024 (ColorMapSize)
#define AniRowIndex(i)		(i << 5)	// Multiply by 32 (rowsize)

/*
 * Enumerated Types
 */

// Yummy Varieties
//
enum {Y_Mult2, Y_Mult3, Y_Mult5, Y_Triple, Y_Uzi, Y_Shield, Y_Creamer, NbrYummies};

// Ship Flags - are we firing?  turning? thrusting? shields up?
//
enum KeyFlags    {KF_Fire=0x01,KF_Left=0x02,KF_Right=0x04,
				  KF_Thrust=0x08,KF_Shield=0x10};

enum ShipFlags	 {SF_Fire=0x01, SF_Left=0x02, SF_Right=0x04,
				  SF_Thrust=0x08, SF_Shield=0x10};

// Sound IDs
//
enum Sounds		 {S_Startup, S_Explosion, S_BadGuyDebut, S_ShipDebut, S_ShipExplodes,
			 	  S_EnemyFires, S_Thrust, S_DudBullet, 
			 	  S_LevelCompletion, S_CompoundFracture, S_Shield, S_YummyConsume,
			 	  S_HighScore1, S_LetterDisplay, S_ExtraShip, S_Loser, S_Firing,
			 	  S_HighScore2, S_JimOuch, S_KevinsPot, S_BreakRings, S_BreakCups,
			 	  S_NbrSounds};


// Game Modes
//
enum GameStates	{ GS_Attract, GS_Play, GS_GameOver, GS_Pause };

// Visual Sprites types
enum SpriteTypes {
				  ST_Wheel, ST_Aster, ST_Teacup, ST_Jim,
				  ST_WheelR, ST_WheelG, ST_WheelB, ST_Aster1, 
				  ST_Aster2, ST_Saucer, ST_Croissant, ST_Cube,
				  ST_Teapot,  ST_TeapotT, ST_TeapotS, ST_TeapotNew,
				  ST_Yummies, ST_Bullet, ST_Photon, ST_Spark,
				  ST_Toaster, ST_SmallCup, ST_Nook, ST_Cranny,
				  ST_BigJim, ST_EMuffin, ST_CoffeePot, ST_CreamBullet,
				  ST_Creamer, ST_Beater, ST_Mixer, ST_Quake,
				  ST_Explosion,
				  NbrSprites};

// Mask which identifies valid targets for "good bullets" (based on sprite type)


// Ship Modes (special guns, magic shield)
enum ShipModes	 {SM_Triple=1, SM_Uzi=2, SM_AutoShield=4, SM_CreamBullet=8};

/*
 * Data Structures
 */

// User Preferences - keyboard remapping, sound level, username
typedef struct {
	char	fireKeyByte, fireKeyBit, fireKeyCode, fireKeyAscii;
	char	thrustKeyByte, thrustKeyBit, thrustKeyCode, thrustKeyAscii;
	char	leftKeyByte, leftKeyBit, leftKeyCode, leftKeyAscii;
	char	rightKeyByte, rightKeyBit, rightKeyCode, rightKeyAscii;
	char	shieldKeyByte, shieldKeyBit, shieldKeyCode, shieldKeyAscii;
	short	soundLevel;
	Str31	userName;
	char	reserved[30];
} PrefsData;

// Sprite Vector Coordinate system uses fixed point arithmetic
//
typedef struct	{
	long	lh,lv;
} LongPoint;

// Data structure for saving the area of screen underneath sprite
//

// Data structure for each sprite instance
//
typedef struct SpriteInstance {
	unsigned long id;
	Point		pos,oldPos;		// Sprite Coords
	LongPoint	vector;			// Sprite Movement Velocity
	short		type;			// Sprite Type - Index into Sprite Def Table
	short		aniState;		// Animation Frame
	short		aniSpeed;		// Animation Frame Rate
	short		tickCtr;		// Time Counter
	short		lifeSpan;		// Life Span of Sprite
	short		angle;			// Index into Vector Table
	short		width;			// Width of Sprite
	unsigned char	active,		// Sprite is Active
					update;		// Sprite needs updating onscreen
	unsigned char maxHits;		// MaxHits
	unsigned char hitsTaken;	// Hits Taken
	unsigned char state;		// Sprite State
	unsigned char stateTimer;	// Sprite State Timer
	unsigned char targetFlags;	// Bullets
	unsigned char savePixel;	// Save pixel
} SpriteInstance;


typedef struct StarInstance {
	Point			pos;
	unsigned char 	color,baseShade,saveColor,flickerCounter;
} StarInstance;

// Sprite Definition Table - defines behavior for each variety of sprite
// function pointers are used for drawing, erasing and moving
//
typedef struct {
	short	type;
	short	class;
	short	firstIconID;
	short	nbrIcons;
	void	(*moveFunc)(register SpriteInstance *sp);
	void	(*eraseFunc)(register SpriteInstance *sp);
	void	(*drawFunc)(register SpriteInstance *sp);
	Ptr		colorMaps;
	Ptr		maskMaps;
	Handle	lSprite;
} SpriteDef;

// Entry in "top 10" scores table
//
typedef struct {
	long	score;
	short	level;
	Str31	name;
} ScoreRecord;

/*
 * Extern Variables - see below for comments
 */

// Connection Stuff
#define ToastEventID	'cTst'

enum	{ C_None, C_AppleTalk, C_IRCTCP, C_IRCSerial, C_NbrConnectionTypes };


#if DEBUGGING
extern Boolean		gDebug;
#endif
extern short		gAssetFile;
extern GWorldPtr	gOffScreen;
extern PixMapHandle	gPixMap;
extern Boolean		g12InchMode,gClientMode;
extern short		gConnectionType;
extern short		gMaxSprite,gSpriteCnt,gSparkCnt,gAsteroidCnt,gMinAsteroids,gRemainingShips;
extern short		gAttractMode;
// extern short		gNbrSaveMaps;
extern short		gYummyCnt,gToasterCount,gMixerCount;
extern long			gScreenRowBytes,gVideoRowBytes;
extern Ptr			gScreenMem,gVideoMem;
extern Rect			gPlayRect,gOffscreenRect,gUnivRect;
extern Point		gCenterP,gVidOffset,gOffsetPt,gOldOffset;

extern long			gPlayWidth,gPlayHeight;
extern short		gVisPlayWidth,gVisPlayHeight,gUnivWidth,gUnivHeight,gGraceTop,gGraceLeft,gGraceRight,gGraceBottom;
extern long			gGameClockTicks,gGameClock;		// 20 Frames per Second
extern long			gLastDispTime;
extern short		gGameState,gGameLevel,gShipMode,gShieldPower,gKeyFlags;
extern long			gGameScore;
extern long			gScoreMultiply;
extern SpriteDef	sDef[NbrSprites];
extern LongPoint	vecTable[MaxAngles];
extern ScoreRecord	scoreRecord[MaxScoreRecords];
extern PrefsData	gPrefs;
// extern SaveMapRecord	*smTable;
extern SpriteInstance	*sTable,*gShip;
extern StarInstance		*starTable;
extern unsigned long	spriteIDCtr;

extern WindowPtr	gWindow;

/*
 * Function Declarations
 */

// Low level Display
void MyCopyBits(void);
void MyCopyBits12Inch(void);
void MyCopyRect(register Rect *r);

// Picture Display
void DisplayPicture(short picID, short hOffset, short vOffset);
void GammaFadeIn(void);
void GammaFadeOut(void);

// Game Control
void InitializeGame(WindowPtr theWin);		// 1 Time only inits
void CleanUp(void);							// 1 Time only cleanup
void MainGameLoop(WindowPtr theWin);
void StartupScreen(void);
void BeginAttract(void);
void BeginGame(void);
void EndGame(void);
void InitStars(void);
void InitLevel(void);
void InitStatusDisplay(void);
void StatusPrintf(short x, short y, char *tmp, ...);
void PrintfXY(short x, short y, char *template, ...);
void CenterString(StringPtr str);
void StatusUpdate();
void AddScore(short amt);
void TrackScore();
void GetUserName(StringPtr name);
void CheckKeys(void);
void LoadPreferences(void);

// Sprite Maintenence
SpriteInstance *NewSprite(void);
void KillSprite(SpriteInstance *sp);
void NewShip(void);
void NewSaucer(void);
void NewCroissant(void);
void NewCube(void);
void NewAsteroid(short type);
void NewStar(void);
void NewYummy(void);
void NewCreamer(void);
void NewToaster(void);
void NewCoffeePot(void);
void LaunchBullet(short type, short x, short y, long vx, long vy, short lifeSpan, short bFlag);
void LaunchSpark(short x, short y, long vx, long vy, short lifeSpan, short color);
void ExplodeSprite(register SpriteInstance *sp, register short lifeSpan);
void LaunchNookandCranny(SpriteInstance *par);

// Util Funcs
short min(register short x, register short y);
short max(register short x, register short y);
void HideMenuBar(WindowPtr theWin);
void RestoreMenuBar();
short MyRandom(short limit);

// Sounds
void InitSounds();
void PlaySound(short i, short priority);
void EndSounds();

// Unsorted
void LoadTopScores();
void DisplayTopScores();
void IntegrateScore(long score, short level);
void EndLevel(void);

#if DEBUGGING
void NewDebugDisplay(void);
#endif

// State Machine Functions


#if DEBSPRITES
void SpriteCntMove(register SpriteInstance *);
void MaxSpriteMove(register SpriteInstance *);
#endif


enum {	SP_Med, SP_Large, SP_Custom };

typedef struct {
	short	width;
	short	height;
	short	nbrFrames;
	short	dataOffset;
	long	frameOffset[1];
	unsigned char data[];
} LSprite;

#define LSPR_Mask	0x80
#define LSPR_Opaque	0x40
#define LSPR_Mixed	0xC0
#define LSPR_LowBits 0x3F
#define LSPR_HiBits 0xC0

#define BF_ShipCollideFlag	0x01
#define BF_GoodBulletFlag	0x02
#define BF_BadBulletFlag	0x03

#define AsterColor	0x13
#define BadColor	0xD8
#define YummyColor	0x92
#define TeaColor	0xC0
#define Invisible	0x00

extern unsigned char gRadarColor[NbrSprites];
extern unsigned char gBulletFlag[NbrSprites];
// The following code is only compiled if "SPRITEMAIN" is turned on
//
#if SPRITEMAIN

#include "CToastVars.h"

#endif

/* End of CToast.h */