// Assets.c		MOSTLY SHARED EXCEPT FILE I/O
//
//				Equivalent of Macintosh Resource Manager used for resources
//				which will be identical across Mac and Windows.
//
//				Currently the only resources types in use are 'Prop' and 'Fave'

// These functions duplicate much of the functionality of the Apple Resource Manager
// but allows for a system where all resources have a longword ID number, which will
// needed on large palace systems with large numbers of props.
//
// once the file handling and memory management is made more portable, this
// should also allow for a somewhat more portable resource management system

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "m-assets.h"

Boolean EqualPString(StringPtr inStr, StringPtr outStr, Boolean caseSens);

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

short			gAssetError;      // Last assets error
AssetFileVars	*firstAssetFile;	// first open resource file/map in linked list
AssetFileVars	*curAssetFile;		// current resource file/map

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr AssetError(void)
{
	return gAssetError;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr NewAssetFileRec(short refNum)
{
	AssetFileVars	*ar;

	gAssetError = 0;

	ar = (AssetFileVars *)NewPtrClear(sizeof(AssetFileVars));
	curAssetFile                = ar;
	ar->nextAssetFile           = firstAssetFile;
	firstAssetFile              = ar;
	ar->aRefNum                 = refNum;
	ar->afHeader.dataOffset     = sizeof(AssetFileHeader);
	ar->afHeader.dataSize       = 0L;
	ar->afHeader.assetMapOffset = sizeof(AssetFileHeader);
	ar->afHeader.assetMapSize   = sizeof(AssetMapHeader);
	ar->typeList                = NewHandleClear(0L);
	ar->assetList               = NewHandleClear(0L);
	ar->nameList                = NewHandleClear(0L);
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr LoadAssetMap(short refNum)
{
	long	          count;
	Handle	        afMap;
	AssetFileVars	 *ar = curAssetFile;
	AssetMapHeader *amh;

	gAssetError = 0;
	count = sizeof(AssetFileHeader);
	if ((gAssetError = FSRead(refNum,&count,(Ptr)&ar->afHeader)) != noErr)
		return gAssetError;

  // esr - check to see if we need to swap bytes
  	ar->afHeader.dataOffset     = MSBToLocalLong(ar->afHeader.dataOffset);
    ar->afHeader.dataSize       = MSBToLocalLong(ar->afHeader.dataSize);
	ar->afHeader.assetMapOffset = MSBToLocalLong(ar->afHeader.assetMapOffset);
  	ar->afHeader.assetMapSize   = MSBToLocalLong(ar->afHeader.assetMapSize);

	afMap = NewHandleClear(ar->afHeader.assetMapSize);
	if ((gAssetError = SetFPos(refNum,fsFromStart,ar->afHeader.assetMapOffset)) != noErr)
		return gAssetError;
	count = ar->afHeader.assetMapSize;
	HLock(afMap);
	if ((gAssetError = FSRead(refNum,&count,*afMap)) != noErr)
		return gAssetError;
	amh = (AssetMapHeader *) *afMap;

  // esr - these need to be swapped too.
  {
    AssetTypeRec   *typePtr;
    AssetRec       *recPtr;
    long            i;

 	  amh->nbrTypes    = MSBToLocalLong(amh->nbrTypes);
 	  amh->nbrAssets   = MSBToLocalLong(amh->nbrAssets);
 	  amh->lenNames    = MSBToLocalLong(amh->lenNames);
 	  amh->typesOffset = MSBToLocalLong(amh->typesOffset);
 	  amh->recsOffset  = MSBToLocalLong(amh->recsOffset);
 	  amh->namesOffset = MSBToLocalLong(amh->namesOffset);


    typePtr = (AssetTypeRec *)(*afMap + amh->typesOffset);
    for (i=0;i<amh->nbrTypes;i++)
    {
    	typePtr->assetType  = MSBToLocalLong(typePtr->assetType);
    	typePtr->nbrAssets  = MSBToLocalLong(typePtr->nbrAssets);
    	typePtr->firstAsset = MSBToLocalLong(typePtr->firstAsset);
      typePtr++;
    }
    recPtr = (AssetRec *)(*afMap + amh->recsOffset);
    for (i=0;i<amh->nbrAssets;i++)
    {
    	recPtr->idNbr       = MSBToLocalLong(recPtr->idNbr);
      recPtr->dataOffset  = MSBToLocalLong(recPtr->dataOffset);
    	recPtr->dataSize    = MSBToLocalLong(recPtr->dataSize);
    	recPtr->lastUseTime = MSBToLocalLong(recPtr->lastUseTime);
    	recPtr->nameOffset  = MSBToLocalLong(recPtr->nameOffset);
    	recPtr->flags       = MSBToLocalLong(recPtr->flags);
    	recPtr->crc    		= MSBToLocalLong(recPtr->crc);		// 6/7/95
      recPtr++;
    }
  }

	SetHandleSize(ar->typeList ,sizeof(AssetTypeRec)*amh->nbrTypes);
	SetHandleSize(ar->assetList,sizeof(AssetRec)*amh->nbrAssets);
	SetHandleSize(ar->nameList ,amh->lenNames);

	BlockMove((Ptr)*afMap,(Ptr)&ar->afMap,sizeof(AssetMapHeader));
	// boom on the following line
	BlockMove((*afMap + amh->typesOffset),*ar->typeList,sizeof(AssetTypeRec)*amh->nbrTypes);
	BlockMove((*afMap + amh->recsOffset) ,*ar->assetList,sizeof(AssetRec)*amh->nbrAssets);
	BlockMove((*afMap + amh->namesOffset),*ar->nameList,amh->lenNames);


	HUnlock(afMap);
	DisposeHandle(afMap);

	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr OpenAssetFile(FSSpec *fsFile,short *refNum)
{
	if ((gAssetError = FSpOpenDF(fsFile,fsRdWrPerm,refNum)) != noErr)
		return gAssetError;
	if ((gAssetError = NewAssetFileRec(*refNum)) != noErr)
		return gAssetError;
	curAssetFile->fsSpec = *fsFile;
	if ((gAssetError = LoadAssetMap(*refNum)) != noErr)
		return gAssetError;

	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr NewAssetFile(FSSpec *fsFile, short *refNum)
{
	gAssetError = 0;
	FSpDelete(fsFile);

	if ((gAssetError = FSpCreate(fsFile,ASSET_CREATOR,ASSET_FILETYPE,0L)) != noErr)
		return gAssetError;
	if ((gAssetError = FSpOpenDF(fsFile,fsRdWrPerm,refNum)) != noErr)
		return gAssetError;
	if ((gAssetError = NewAssetFileRec(*refNum)) != noErr)
		return gAssetError;
	curAssetFile->fsSpec = *fsFile;
	curAssetFile->fileNeedsUpdate = true;
	if ((gAssetError = UpdateAssetFile()) != noErr)
		return gAssetError;
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr WriteAsset(AssetRec *ap)
{
	AssetFileVars	*ar = curAssetFile;
	long	count;
	char	hState;
	short	oRefNum = curAssetFile->aRefNum;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return gAssetError;
	}
	ap->flags &= ~AssetModified;
	count = ap->dataSize;
	hState = HGetState(ap->rHandle);
	HLock(ap->rHandle);
	ap->dataOffset = ar->afHeader.dataSize;
	if ((gAssetError = SetFPos(oRefNum,fsFromStart,ap->dataOffset + ar->afHeader.dataOffset)) != noErr)
		return gAssetError;
	if ((gAssetError = FSWrite(oRefNum,&count,*ap->rHandle)) != noErr)
		return gAssetError;
	ar->afHeader.dataSize += ap->dataSize;
	HSetState(ap->rHandle,hState);
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr	WriteAssetMap(short oRefNum)
{
	AssetFileVars	*ar = curAssetFile;
	AssetRec		*ap;
	long			count;
	Handle			tempAssetList;
	long			i;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return gAssetError;
	}

	// Write Asset Map
	ar->afMap.typesOffset = sizeof(AssetMapHeader);
	ar->afMap.recsOffset = ar->afMap.typesOffset + ar->afMap.nbrTypes*sizeof(AssetTypeRec);
	ar->afMap.namesOffset = ar->afMap.recsOffset + ar->afMap.nbrAssets*sizeof(AssetRec);

	ar->afHeader.assetMapOffset = ar->afHeader.dataOffset + ar->afHeader.dataSize;
	ar->afHeader.assetMapSize = ar->afMap.namesOffset + ar->afMap.lenNames;

	if ((gAssetError = SetFPos(oRefNum,fsFromStart,ar->afHeader.assetMapOffset)) != noErr)
		return gAssetError;

	count = sizeof(AssetMapHeader);
	gAssetError = FSWrite(oRefNum,&count,(Ptr)&ar->afMap);
    if (gAssetError != noErr)
		return gAssetError;
	HLock(ar->typeList);
	count = ar->afMap.nbrTypes*sizeof(AssetTypeRec);
	if ((gAssetError = FSWrite(oRefNum,&count,*ar->typeList)) != noErr)
		return gAssetError;
	HUnlock(ar->typeList);

	// Make copy of AssetList, so we can save with correct flags
	tempAssetList = ar->assetList;
	HandToHand(&tempAssetList);
	ap = (AssetRec *) *tempAssetList;
	// Write Data for Each Asset, (clear modified flag first)
	for (i = 0; i < ar->afMap.nbrAssets; ++i,++ap)
		ap->flags &= ~AssetLoaded;

	HLock(tempAssetList);
	count = ar->afMap.nbrAssets*sizeof(AssetRec);
	if ((gAssetError = FSWrite(oRefNum,&count,*tempAssetList)) != noErr)
		return gAssetError;
	HUnlock(tempAssetList);
	DisposeHandle(tempAssetList);

	HLock(ar->nameList);
	count = ar->afMap.lenNames;
	if ((gAssetError = FSWrite(oRefNum,&count,*ar->nameList)) != noErr)
		return gAssetError;
	HUnlock(ar->nameList);

	// Update Header
	if ((gAssetError = SetFPos(oRefNum,fsFromStart,0L)) != noErr)
		return gAssetError;
	count = sizeof(AssetFileHeader);
	if ((gAssetError = FSWrite(oRefNum,&count,(Ptr)&ar->afHeader)) != noErr)
		return gAssetError;
#if macintosh			// 6/24/95 Flush the disc cache to reduce chances of corruption
	FlushVol(NULL, ar->fsSpec.vRefNum);
#endif
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Keep tweaking filename until it's UNIQUE
#if macintosh
void InsureNameIsUnique(FSSpec *tSpec);
void InsureNameIsUnique(FSSpec *tSpec)
{
	FInfo			fInfo;
	short			MyRandom(short max);
	while (HGetFInfo(tSpec->vRefNum,
					 tSpec->parID,
					 tSpec->name, &fInfo) == noErr) {
		tSpec->name[1+MyRandom(tSpec->name[0])] = 'a'+MyRandom(26);
	}
}
#endif

// Write out entire Asset File from Scratch (this compresses out old assets)
// 6/24/95 Modified to write file in place, so that file refNum isn't changed
//
// 7/10/95 Fixed bug which caused file corruption if any assets were currently loaded
//
ErrNbr	WriteAssetFile(void)
{
	AssetFileVars	*ar = curAssetFile;
	AssetRec		*ap;
	short			astRefNum = curAssetFile->aRefNum,tempRefNum;
	long			count;
	long			i;
	Handle			th;
	char			hState;
  #ifdef macintosh
	FSSpec			tempSpec = {0,0L,"\pTEMP.AST"};
  #else
	FSSpec			tempSpec = {0,0L,"TEMP.AST"};
  #endif

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return gAssetError;
	}

	// Open a temporary file for writing
#ifdef macintosh
	// !!! This realy should be done on both platforms!  
	InsureNameIsUnique(&tempSpec);
#else
	FSpDelete(&tempSpec);
#endif
	if ((gAssetError = FSpCreate(&tempSpec,ASSET_CREATOR,ASSET_FILETYPE,0L)) != noErr)
		return gAssetError;
	if ((gAssetError = FSpOpenDF(&tempSpec,fsRdWrPerm,&tempRefNum)) != noErr)
		return gAssetError;

	// Write Header
	count = sizeof(AssetFileHeader);
	if ((gAssetError = FSWrite(tempRefNum,&count,(Ptr)&ar->afHeader)) != noErr)
		return gAssetError;
	ar->afHeader.dataSize = 0L;
	ar->afHeader.dataOffset = sizeof(AssetFileHeader);

	HLock(ar->assetList);
	ap = (AssetRec *) *ar->assetList;

	// Write Data for Each Asset, (clear modified flag first)
	for (i = 0; i < ar->afMap.nbrAssets; ++i,++ap) {
		count = ap->dataSize;

		if (ap->flags & AssetLoaded) {
			// no no no - don't write into old file!  7/10/95
			//
			// if ((gAssetError = WriteAsset(ap)) != noErr)
			//	return gAssetError;
			th = ap->rHandle;
			hState = HGetState(th);
			HLock(th);
		}
		/*** else { ***/		// 7/10/95 bugfix, copy out loaded assets as well...
		else {
			SetFPos(astRefNum,fsFromStart,ap->dataOffset+ar->afHeader.dataOffset);
			th = NewHandle(ap->dataSize);
			HLock(th);
			if ((gAssetError = FSRead(astRefNum,&count,*th)) != noErr)
				return gAssetError;
		}
		ap->dataOffset = ar->afHeader.dataSize;
		if ((gAssetError = FSWrite(tempRefNum,&count,*th)) != noErr)
			return gAssetError;
		ar->afHeader.dataSize += ap->dataSize;
		
		// 7/10/95 Only delete handle if asset isn't loaded, otherwise, unlock it
		if (!(ap->flags & AssetLoaded))
			DisposeHandle(th);
		else
			HSetState(th,hState);
	}
	HUnlock(ar->assetList);
	// 6/24/95 NEW NEW NEW
	// Copy contents of Temp File (tempRefNum) into Our File (astRefNum)
	//
	{
#define BUFSIZE	32000L
		long	amt = ar->afHeader.dataSize + ar->afHeader.dataOffset;
		long	count;
		Ptr		tbuf;		
		SetFPos(tempRefNum,fsFromStart,0L);
		SetFPos(astRefNum,fsFromStart,0L);
		tbuf = NewPtr(BUFSIZE);
		while (amt) {
			if (amt > BUFSIZE)
				count = BUFSIZE;
			else
				count = amt;
			if ((gAssetError = FSRead(tempRefNum,&count,tbuf)) != noErr)
				return gAssetError;
			if ((gAssetError = FSWrite(astRefNum,&count,tbuf)) != noErr)
				return gAssetError;
			amt -= count;
		}
	}
	FSClose(tempRefNum);
	FSpDelete(&tempSpec);
	
	if ((gAssetError = WriteAssetMap(astRefNum)) != noErr)
		return gAssetError;

	// Reduce length of file...
	SetEOF(astRefNum,ar->afHeader.assetMapOffset + ar->afHeader.assetMapSize);
#if macintosh			// 6/24/95 Flush the disc cache to reduce chances of corruption
	FlushVol(NULL, ar->fsSpec.vRefNum);
#endif

/******
	// Close & Delete Old File
	// !! Use correct directory info...
	FSClose(astRefNum);
	FSpDelete(&ar->fsSpec);

	// Close New File
	FSClose(tempRefNum);

	// Rename Temp File, reset file pointers
	if ((gAssetError = FSpRename(&tempSpec,ar->fsSpec.name)) != noErr)
		return gAssetError;

	ar->fsSpec = tempSpec;

	// Reopen New File
	if ((gAssetError = FSpOpenDF(&ar->fsSpec,fsRdWrPerm,&curAssetFile->aRefNum)) != noErr)
		return gAssetError;

*******/
	ar->fileNeedsUpdate = false;
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Write out changed assets, and new asset map & header
// More Efficient, but will leave deleted assets in file
//
ErrNbr	UpdateAssetFile(void)
{
	short			oRefNum = curAssetFile->aRefNum;
	AssetFileVars	*ar = curAssetFile;
	AssetRec		*ap;
	long			count;
	long			i;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return gAssetError;
	}
	if (!ar->fileNeedsUpdate)
		return noErr;

	// Update Header
	if ((gAssetError = SetFPos(oRefNum,fsFromStart,0L)) != noErr)
		return gAssetError;
	count = sizeof(AssetFileHeader);
	if ((gAssetError = FSWrite(oRefNum,&count,(Ptr)&ar->afHeader)) != noErr)
		return gAssetError;
	if ((gAssetError = SetFPos(oRefNum,fsFromStart,ar->afHeader.dataOffset+ar->afHeader.dataSize)) != noErr)
		return gAssetError;
	// Write Data for Each Asset, (clear modified flag first)
	HLock(ar->assetList);
	ap = (AssetRec *) *ar->assetList;
	for (i = 0; i < ar->afMap.nbrAssets; ++i,++ap) {
		if (ap->flags & AssetModified) {
			if ((gAssetError = WriteAsset(ap)) != noErr)
				return gAssetError;
		}
	}
	HUnlock(ar->assetList);
	if ((gAssetError = WriteAssetMap(oRefNum)) != noErr)
		return gAssetError;
	ar->fileNeedsUpdate = false;
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

short	CurAssetFile(void)
{
	gAssetError = 0;
	if (curAssetFile)
		return curAssetFile->aRefNum;
	else
		return 0;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr UseAssetFile(short refNum)
{
	AssetFileVars	*ar;

	gAssetError = 0;

	ar = firstAssetFile;
	while (ar && ar->aRefNum != refNum)
		ar = ar->nextAssetFile;
	if (ar) {
		curAssetFile = ar;
		return noErr;
	}
	else {
		gAssetError = -1;
		return gAssetError;
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr CloseAssetFile(short refNum)
{
	AssetFileVars	*ar,*tr;
	AssetRec	  	*ap;
	short		     	 sr;
	short			     i;

	gAssetError = 0;
	if (curAssetFile == NULL)
    return gAssetError = -1;
	if (refNum == curAssetFile->aRefNum)
  {
		if (curAssetFile->nextAssetFile)
			sr = curAssetFile->nextAssetFile->aRefNum;
		else
			sr = 0;
	}
	else
  {
		sr = CurAssetFile();
		UseAssetFile(refNum);
	}
	ar = curAssetFile;
	UpdateAssetFile();
	FSClose(ar->aRefNum);
	if (ar == firstAssetFile)
		firstAssetFile = ar->nextAssetFile;
	else
  {
		tr = firstAssetFile;
		while (tr && tr->nextAssetFile != ar)
			tr = tr->nextAssetFile;
		tr->nextAssetFile = ar->nextAssetFile;
	}
	// !! Dispose of Handles
	HLock(ar->assetList);
	ap = (AssetRec *) *ar->assetList;
	for (i = 0; i < ar->afMap.nbrAssets; ++i,++ap)
		if (ap->flags & AssetLoaded)
			DisposeHandle(ap->rHandle);
	HUnlock(ar->assetList);
	DisposeHandle(ar->typeList);
	DisposeHandle(ar->assetList);
	DisposeHandle(ar->nameList);
	DisposePtr((Ptr) ar);
	if (sr)
		UseAssetFile(sr);
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

AssetRec	*FindAssetRec(AssetType type, long id)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at;
	AssetRec		*ap;
	long			i,j,nbrTypes;
	at = (AssetTypeRec *) *ar->typeList;
	nbrTypes = ar->afMap.nbrTypes;
	for (i = 0; i < nbrTypes; ++i,++at) {
	if (at->assetType == MSBToLocalLong(type)) //esr -
    {
			ap = (AssetRec *) *ar->assetList;
			ap += at->firstAsset;
			for (j = at->nbrAssets-1; j >= 0; --j,++ap) {
				if (ap->idNbr == id)
					return ap;
			}
			break;
		}
	}
	return NULL;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

long UniqueAssetID(AssetType type, long minID)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at;
	AssetRec		*ap;
	long			i,j,nbrTypes,uniqueID,nbrAssets;
	Boolean			uniqueFlag;

	at = (AssetTypeRec *) *ar->typeList;
	nbrTypes = ar->afMap.nbrTypes;
	uniqueID = minID;
	for (i = 0; i < nbrTypes; ++i,++at) {
		if (at->assetType == type) {
			do {
				uniqueFlag = true;
				ap = (AssetRec *) *ar->assetList;
				ap += at->firstAsset;	
				nbrAssets = at->nbrAssets;
				for (j = 0; j < nbrAssets; ++j,++ap) {
					if (ap->idNbr == uniqueID) {
						uniqueFlag = false;
						++uniqueID;
						break;
					}
				}
			} while (!uniqueFlag);
			break;
		}
	}
	return uniqueID;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

AssetRec	*FindAssetRecByIndex(AssetType type, long idx)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at;
	AssetRec		*ap;
	long			i,j,nbrAssets,nbrTypes;
	at = (AssetTypeRec *) *ar->typeList;
	nbrTypes = ar->afMap.nbrTypes;
	for (i = 0; i < nbrTypes; ++i,++at) {
		if (at->assetType == type) {
			ap = (AssetRec *) *ar->assetList;
			ap += at->firstAsset;	
			nbrAssets = at->nbrAssets;
			for (j = 0; j < nbrAssets; ++j,++ap) {
				if (j == idx)
					return ap;
			}
			break;
		}
	}
	return NULL;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Handle	GetAsset(AssetType type, long id)
{
	AssetFileVars	*ar = curAssetFile;
	AssetRec		*ap;
	Handle			h;
	long			count;
	unsigned long	t;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return NULL;
	}
	ap = FindAssetRec(type,id);
	if (ap == NULL) {
		gAssetError = resNotFound;
		return NULL;
	}
	GetDateTime(&t);
	ap->lastUseTime = t;
	if (ap->flags & AssetLoaded)
		return ap->rHandle;
	else {
		HLock(ar->typeList);
		HLock(ar->assetList);
		h = NewHandle(ap->dataSize);
		HUnlock(ar->typeList);
		HUnlock(ar->assetList);
		HLock(h);
		SetFPos(ar->aRefNum,fsFromStart,ap->dataOffset+ar->afHeader.dataOffset);
		count = ap->dataSize;
		FSRead(ar->aRefNum,&count,*h);
		HUnlock(h);
		ap->rHandle = h;
		ap->flags |= AssetLoaded;
		return ap->rHandle;
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr	ChangedAsset(Handle h)
{
	AssetFileVars	*ar = curAssetFile;
	AssetRec	*ap = (AssetRec *) *curAssetFile->assetList;
	long	i,nbrAssets = curAssetFile->afMap.nbrAssets;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return gAssetError;
	}
	for (i = 0; i < nbrAssets; ++i,++ap) {
		if ((ap->flags & AssetLoaded) && ap->rHandle == h) {
			ap->flags |= AssetModified;
			ap->dataSize = GetHandleSize(h);
			if (ap->dataSize == 332 && ap->idNbr < 0)
				DebugStr("\pOy!");
			// ap->crc = ComputeHandleCRC(ap->rHandle,ap->dataSize);  6/9/95
			ar->fileNeedsUpdate = true;
		}
	}
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Boolean FindAssetByHandle(Handle rHandle, AssetTypeRec **att, AssetRec **app)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at,*fat;
	AssetRec		*ap;
	long				i,nbrAssets;

	ap = (AssetRec *) *ar->assetList;
	at = (AssetTypeRec *) *ar->typeList;
	nbrAssets = ar->afMap.nbrAssets;
	for (i = 0; i < nbrAssets; ++i,++ap) {
		if (i == at->firstAsset) {
			fat = at;
			++at;
		}
		if ((ap->flags & AssetLoaded) && ap->rHandle == rHandle) {
			*att = fat;
			*app = ap;
			return true;
		}
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Boolean AssetExists(AssetType type, long id)
{
	return FindAssetRec(type,id) != NULL;
}

Boolean AssetExistsWithCRC(AssetType type, long id, unsigned long crc)
{
	return FindAssetRecWithCRC(type,id,crc) != NULL;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr	GetAssetInfo(Handle h, AssetType *type, long *id, StringPtr name)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at;
	AssetRec		*ap;
	StringPtr		as;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return gAssetError;
	}
	if (!FindAssetByHandle(h,&at,&ap)) {
		gAssetError = resNotFound;
		return gAssetError;
	}
	*type = at->assetType;
	*id = ap->idNbr;
	if (ap->nameOffset != -1) {
		as = (StringPtr) (*ar->nameList + ap->nameOffset);
		BlockMove(as,name,as[0]+1);
	}
	else
		name[0] = 0;
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Fix to allow changes of type
ErrNbr	SetAssetInfo(Handle h, long id, StringPtr name)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at;
	AssetRec		*ap;
	StringPtr		as;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return gAssetError;
	}
	if (!FindAssetByHandle(h,&at,&ap)) {
		gAssetError = resNotFound;
		return gAssetError;
	}
	ap->idNbr = id;
	if (name[0]) {
		if (ap->nameOffset != -1) {
			as = (StringPtr) (*ar->nameList + ap->nameOffset);
			if (as[0] >= name[0]) {
				BlockMove(name,as,name[0]+1);
			}
			else {
				// New String
				SetHandleSize(ar->nameList,ar->afMap.lenNames+name[0]+1);
				ap->nameOffset = ar->afMap.lenNames;
				as = (StringPtr) (*ar->nameList + ap->nameOffset);
				BlockMove(name,as,name[0]+1);
				ar->afMap.lenNames += name[0]+1;
			}
		}
		else {
			// New String
			SetHandleSize(ar->nameList,ar->afMap.lenNames+name[0]+1);
			ap->nameOffset = ar->afMap.lenNames;
			as = (StringPtr) (*ar->nameList + ap->nameOffset);
			BlockMove(name,as,name[0]+1);
			ar->afMap.lenNames += name[0]+1;
		}
	}
	else
		ap->nameOffset = -1;
	ar->fileNeedsUpdate = true;
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

long	CountAssets(AssetType type)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at;
	long			i,nbrTypes;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return gAssetError;
	}
	at = (AssetTypeRec *) *ar->typeList;
	nbrTypes = ar->afMap.nbrTypes;
	for (i = 0; i < nbrTypes; ++i,++at) {
		if (at->assetType == type) {
			return at->nbrAssets;
		}
	}
	return 0L;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Handle	GetIndAsset(AssetType type, long nbr)
{
	AssetFileVars	*ar = curAssetFile;
	AssetRec		*ap;
	Handle			h;
	long			count;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return NULL;
	}
	ap = FindAssetRecByIndex(type,nbr);
	if (ap == NULL) {
		gAssetError = resNotFound;
		return NULL;
	}
	if (ap->flags & AssetLoaded)
		return ap->rHandle;
	else {
		HLock(ar->typeList);
		HLock(ar->assetList);
		h = NewHandle(ap->dataSize);
		HUnlock(ar->typeList);
		HUnlock(ar->assetList);
		HLock(h);
		SetFPos(ar->aRefNum,fsFromStart,ap->dataOffset+ar->afHeader.dataOffset);
		count = ap->dataSize;
		FSRead(ar->aRefNum,&count,*h);
		HUnlock(h);
		ap->rHandle = h;
		ap->flags |= AssetLoaded;
		return ap->rHandle;
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

long	CountAssetTypes()
{
	AssetFileVars	*ar = curAssetFile;
	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return gAssetError;
	}
	return ar->afMap.nbrTypes;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

AssetType GetIndAssetType(long nbr)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at;
	long			i,nbrTypes;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return gAssetError;
	}
	at = (AssetTypeRec *) *ar->typeList;
	nbrTypes = ar->afMap.nbrTypes;
	for (i = 0; i < nbrTypes; ++i,++at) {
		if (i == nbr)
			return at->assetType;
	}
	return 0L;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr	AddAsset(Handle h, AssetType type, long id, StringPtr name)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at;
	AssetRec		*ap;
	long			i,nbrTypes,typeNbr;
	unsigned long	t,curNbr,desNbr;
	StringPtr		as;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return gAssetError;
	}

	// Find if Type already exists
	at = (AssetTypeRec *) *ar->typeList;
	nbrTypes = ar->afMap.nbrTypes;
	for (i = 0; i < nbrTypes; ++i,++at) {
		if (at->assetType == type)
			break;
	}
	typeNbr = i;

	// If not, add & initialize new type record
	if (i >= nbrTypes) {
		ar->afMap.nbrTypes++;
		SetHandleSize(ar->typeList,ar->afMap.nbrTypes * sizeof(AssetTypeRec));
		at = (AssetTypeRec *) *ar->typeList;
		at += ar->afMap.nbrTypes-1;
		at->assetType = type;
		at->nbrAssets = 0;
		at->firstAsset = ar->afMap.nbrAssets;
	}
	HLock(ar->typeList);

	// Add new Resource Record
	++ar->afMap.nbrAssets;
	SetHandleSize(ar->assetList,ar->afMap.nbrAssets * sizeof(AssetRec));
	ap = (AssetRec *) *ar->assetList;
	ap += ar->afMap.nbrAssets-1;
	HLock(ar->assetList);

	// Initialize it
	ap->idNbr = id;
	ap->rHandle = h;
	ap->dataOffset = 0;
	ap->flags = AssetLoaded | AssetModified;
	ap->dataSize = GetHandleSize(h);
	GetDateTime(&t);
	ap->lastUseTime = t;
	// ap->crc = ComputeHandleCRC(h,ap->dataSize);	JAB 6/9/95
	if (name && name[0]) {
		// New String
		SetHandleSize(ar->nameList,ar->afMap.lenNames+name[0]+1);
		ap->nameOffset = ar->afMap.lenNames;
		as = (StringPtr) (*ar->nameList + ap->nameOffset);
		BlockMove(name,as,name[0]+1);
		ar->afMap.lenNames += name[0]+1;
	}
	else
		ap->nameOffset = -1L;
	// ap->reserved = 0L;

	// Swap it into place, updating relevant type records (firstAsset) field
	curNbr = ar->afMap.nbrAssets-1;
	desNbr = at->firstAsset + at->nbrAssets;
	if (curNbr != desNbr) {
		AssetRec	*app,tempRec;
		tempRec = *ap;
		app = ap;
		while (curNbr > desNbr) {
			*app = *(app-1);
			--app;
			--curNbr;
		}
		*app = tempRec;
	}

	// Update Type Record
	at->nbrAssets++;

	// Update firstAsset address of remaining type records
	++typeNbr;
	++at;
	while (typeNbr < nbrTypes) {
		at->firstAsset++;
		++at;
		++typeNbr;
	}
	HUnlock(ar->typeList);
	HUnlock(ar->assetList);
	ar->fileNeedsUpdate = true;
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr RmveAsset(Handle rHandle)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at,*fat=NULL;
	AssetRec		*ap;
	long			nbrAssets,nbrTypes,i,j;
	
	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return gAssetError;
	}
	ap = (AssetRec *) *ar->assetList;
	at = (AssetTypeRec *) *ar->typeList;
	nbrAssets = ar->afMap.nbrAssets;
	nbrTypes = ar->afMap.nbrTypes;
	j = -1;
	for (i = 0; i < nbrAssets; ++i,++ap) {
		if (i == at->firstAsset) {
			fat = at;
			++j;
			++at;
		}
		if ((ap->flags & AssetLoaded) && ap->rHandle == rHandle) {

			// Got it!
			//
			// Dispose of Asset Record
			//
			if (i < nbrAssets-1)		// 6/9/95
				BlockMove((Ptr)(ap+1),(Ptr)ap,(nbrAssets-(i+1))*sizeof(AssetRec));	
			ar->afMap.nbrAssets--;

			// Update Asset Type
			//
			fat->nbrAssets--;

			// Dispose of Asset Type, possibly
			//
			if (fat->nbrAssets <= 0) {
				if (j < nbrTypes-1)		// 6/9/95 bug fix
					BlockMove((Ptr)(fat+1),(Ptr)fat,(nbrTypes-(j+1))*sizeof(AssetTypeRec));
				ar->afMap.nbrTypes--;
				--at;
			}
			else
				++j;

			// Update Remaining Type Records
			while (j < ar->afMap.nbrTypes) {
				at->firstAsset--;
				++j;
				++at;
			}

			// Tag the file for update
			//
			ar->fileNeedsUpdate = true;

			return noErr;
		}
	}
	gAssetError = resNotFound;
	return gAssetError;
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr DetachAsset(Handle h)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at;
	AssetRec		  *ap;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return gAssetError;
	}
	if (!FindAssetByHandle(h,&at,&ap)) {
		gAssetError = resNotFound;
		return gAssetError;
	}
	if (ap->flags & AssetModified) {
		WriteAsset(ap);
	}
	ap->rHandle = NULL;
	ap->flags &= ~AssetLoaded;
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ErrNbr ReleaseAsset(Handle h)
{
	AssetRec		  *ap;
	AssetTypeRec	*at;
	AssetFileVars	*ar = curAssetFile;

	gAssetError = 0;
	if (ar == NULL)
		return (gAssetError = -1);

	if (!FindAssetByHandle(h,&at,&ap))
		return (gAssetError = resNotFound);

	if (ap->flags & AssetModified)
		WriteAsset(ap);

	DisposeHandle(ap->rHandle);
	ap->rHandle = NULL;
	ap->flags  &= ~AssetLoaded;
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 6/7/95 JAB
#define CRC_MAGIC	0xd9216290L
unsigned long ComputeCRC(Ptr pt, long len)
{
	unsigned long crc = CRC_MAGIC;
	unsigned char *p = (unsigned char *) pt;
	while (len--)
		crc = ((crc << 1L) | ((crc & 0x80000000L)? 1 : 0)) ^ *(p++);
	return crc;
}

// 6/7/95 JAB - Get Asset Record that matches ID and CRC
AssetRec	*FindAssetRecWithCRC(AssetType type, long id, unsigned long crc)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at;
	AssetRec		*ap;
	long			i,j,nbrTypes;
	at = (AssetTypeRec *) *ar->typeList;
	nbrTypes = ar->afMap.nbrTypes;
	for (i = 0; i < nbrTypes; ++i,++at) {
	if (at->assetType == MSBToLocalLong(type)) //esr -
    {
			ap = (AssetRec *) *ar->assetList;
			ap += at->firstAsset;
			for (j = at->nbrAssets-1; j >= 0; --j,++ap) {
				if (ap->idNbr == id && (ap->crc == crc || crc == 0L))
					return ap;
			}
			break;
		}
	}
	return NULL;
}

// 6/7/95 JAB - Get Asset that matches ID and CRC
//            - If CRC is zero, it acts like GetAsset
Handle	GetAssetWithCRC(AssetType type, long id, unsigned long crc)
{
	AssetFileVars	*ar = curAssetFile;
	AssetRec		*ap;
	Handle			h;
	long			count;
	unsigned long	t;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return NULL;
	}
	ap = FindAssetRecWithCRC(type,id,crc);
	if (ap == NULL) {
		gAssetError = resNotFound;
		return NULL;
	}
	GetDateTime(&t);
	ap->lastUseTime = t;
	if (ap->flags & AssetLoaded)
		return ap->rHandle;
	else {
		HLock(ar->typeList);
		HLock(ar->assetList);
		h = NewHandle(ap->dataSize);
		HUnlock(ar->typeList);
		HUnlock(ar->assetList);
		HLock(h);
		SetFPos(ar->aRefNum,fsFromStart,ap->dataOffset+ar->afHeader.dataOffset);
		count = ap->dataSize;
		FSRead(ar->aRefNum,&count,*h);
		HUnlock(h);
		ap->rHandle = h;
		ap->flags |= AssetLoaded;
		return ap->rHandle;
	}
}

// 6/7/95
unsigned long GetAssetCRC(Handle h)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at;
	AssetRec		*ap;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return 0L;
	}
	if (!FindAssetByHandle(h,&at,&ap)) {
		gAssetError = resNotFound;
		return 0L;
	}
	return ap->crc;
}

ErrNbr SetAssetCRC(Handle h, unsigned long crc)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at;
	AssetRec		*ap;

	gAssetError = 0;
	if (ar == NULL)
		return gAssetError = -1;
	if (!FindAssetByHandle(h,&at,&ap))
		return gAssetError = resNotFound;
	ap->crc = crc;
	ar->fileNeedsUpdate = true;
	return noErr;
}

AssetRec	*FindAssetRecByName(AssetType type, StringPtr name)
{
	AssetFileVars	*ar = curAssetFile;
	AssetTypeRec	*at;
	AssetRec		*ap;
	long			i,j,nbrTypes;

	if (name == NULL || name[0] == 0)
		return NULL;

	at = (AssetTypeRec *) *ar->typeList;
	nbrTypes = ar->afMap.nbrTypes;
	for (i = 0; i < nbrTypes; ++i,++at) {
	if (at->assetType == MSBToLocalLong(type)) //esr -
    {
			ap = (AssetRec *) *ar->assetList;
			ap += at->firstAsset;
			for (j = at->nbrAssets-1; j >= 0; --j,++ap) {
				if (ap->nameOffset >= 0 && EqualPString(name,(StringPtr) (*ar->nameList + ap->nameOffset),false))
					return ap;
			}
			break;
		}
	}
	return NULL;
}

Handle	GetAssetByName(AssetType type, StringPtr name)
{
	AssetFileVars	*ar = curAssetFile;
	AssetRec		*ap;
	Handle			h;
	long			count;
	unsigned long	t;

	gAssetError = 0;
	if (ar == NULL) {
		gAssetError = -1;
		return NULL;
	}
	ap = FindAssetRecByName(type,name);
	if (ap == NULL) {
		gAssetError = resNotFound;
		return NULL;
	}
	GetDateTime(&t);
	ap->lastUseTime = t;
	if (ap->flags & AssetLoaded)
		return ap->rHandle;
	else {
		HLock(ar->typeList);
		HLock(ar->assetList);
		h = NewHandle(ap->dataSize);
		HUnlock(ar->typeList);
		HUnlock(ar->assetList);
		HLock(h);
		SetFPos(ar->aRefNum,fsFromStart,ap->dataOffset+ar->afHeader.dataOffset);
		count = ap->dataSize;
		FSRead(ar->aRefNum,&count,*h);
		HUnlock(h);
		ap->rHandle = h;
		ap->flags |= AssetLoaded;
		return ap->rHandle;
	}
}
